//---------------------------------------------------------------------------

/*

Robot rebooter if JBT computer down.
MS Visual C++ 6 SP1.
hz37 10022001.

*/

//---------------------------------------------------------------------------

#include "stdafx.h"

//---------------------------------------------------------------------------

// ICMP header.

typedef struct
{
  BYTE   Type;
  BYTE   Code;
  USHORT Checksum;
  USHORT ID;
  USHORT Sequence;
  ULONG  TimeStamp;
} TICMPHeader;

typedef enum 
{
	emInvalid, 
	emNeutral, 
	emArmedAndDangerous, 
	emFinished
} TMode;

//---------------------------------------------------------------------------

// App name for this application (arbitrary stuff).
static const char CAppName[]         = "jultbot";

// How often we check if we're still online (sec).
// After check has failed, check again after CReCheckInterval (sec).
#ifdef NODEBUG
static const int CCheckInterval      = 600;
#else
static const int CCheckInterval      = 30;
#endif
static const int CReCheckInterval    = 10;

// Default size for TCP/IP packet.
static const int CDefPacketSize      = 16;

// Januari 1, 2001, 00:00:00" as a default
// when no last reboot could be read from file.
static const __int64 CDefSysTime     = 126227808000000000;

// Header constant for ICMP type.
static const BYTE CICMPEcho          = 8;

// File to save last reboot systemtime to.
static const char CFileName[]        = "c:\\jultbot.bin";

// Maximal size for TCP/IP packet.
static const int CMaxPacket          = 1024;

// Minimal time between reboots (100 nanosecs units).
#ifdef NODEBUG
static const LONGLONG CMinRebootDiff = 70000000000;
#else
static const LONGLONG CMinRebootDiff =  1200000000; // 2 mins.
#endif

// What address to ping (dotted to avoid DNS lookup).
#ifdef NODEBUG
static const char CTargetAddress[]   = "195.190.241.33";
#else
static const char CTargetAddress[]   = "193.148.15.48"; // JBT tipped 10022001.
#endif

// Winsock timeout in msec (after which it's assumed we're offline).
static const int CSocketTimeout      = 800;

// Interval for WM_TIMER timer (msec).
static const UINT CTimerInterval     = 1000;

//---------------------------------------------------------------------------

// Poll last reboot time at program start (read from file).
LARGE_INTEGER MLastReboot;

// MMode indicates the current mode of the program.
// If no connection exists after the long interval expires,
// MMode goes to emArmedAndDangerous (still no connection)
// and then after a short interval, the relais is turned on if there
// is still no connection.

TMode MMode = emNeutral;

//---------------------------------------------------------------------------

USHORT Checksum(USHORT*, const int);
void DoCheck(TMode&);
void FillIcmpData(char*, const int);
void GetLastReboot(LARGE_INTEGER&);
bool IsOnline();
void KillTasks();
LRESULT CALLBACK MessageHandler(HWND, UINT, WPARAM, LPARAM);
bool RecentlyRebooted();
void Relais(const bool On = true);

#ifndef NODEBUG
void ShowStatus(HWND, const TMode, const int, const int);
#endif

void UpdateLastReboot(LARGE_INTEGER&);

//---------------------------------------------------------------------------

USHORT Checksum(USHORT* Buffer, const int BufferSize)
{

	unsigned long ReturnValue = 0;
	int DataSize = BufferSize;

	while(DataSize > 1)
	{
		ReturnValue += *Buffer++;
		DataSize -= sizeof(USHORT);
	}
  
	if(DataSize)
	{
		ReturnValue += *(UCHAR*)Buffer;
	}

	ReturnValue = (ReturnValue >> 16) + (ReturnValue & 0xffff);
	ReturnValue += (ReturnValue >> 16);

	return static_cast <USHORT> (~ReturnValue);

}

//---------------------------------------------------------------------------

// Check if we're online, unless last reboot was too short time ago.
// Input parameter is changed if needed.

void DoCheck(TMode& CurrentMode)
{

	if(RecentlyRebooted() || IsOnline())
	{
		// No matter what CurrentMode was on input,
		// switch back to neutral and return.
		// Either everything is cool, or we have nothing to do yet.
		CurrentMode = emNeutral;

		return;
	}

	// ******************
	// We are not online!
	// ******************

	switch(CurrentMode)
	{
		case emNeutral:
			CurrentMode = emArmedAndDangerous;
			break;

		case emArmedAndDangerous:
			// The second test has failed, reboot.
			CurrentMode = emFinished;
			UpdateLastReboot(MLastReboot);

			KillTasks();

			Sleep(1000);

			Relais();

		// default case skipped on purpose.
	}

}

//---------------------------------------------------------------------------

// Helper function to fill in various stuff in our ICMP request.

void FillIcmpData(char* ICMPData, const int DataSize)
{

	TICMPHeader* ICMPHeader = reinterpret_cast <TICMPHeader*> (ICMPData);

	ICMPHeader->Type     = CICMPEcho;
	ICMPHeader->Code     = 0;
	ICMPHeader->ID       = static_cast <USHORT> (GetCurrentProcessId());
	ICMPHeader->Checksum = 0;
	ICMPHeader->Sequence = 0;
  
	char* DataPart = ICMPData + sizeof(TICMPHeader);

	// Place some junk in the buffer.
	memset(DataPart, 'E', DataSize - sizeof(TICMPHeader));

}

//---------------------------------------------------------------------------

// Read last reboot date/time from file 
// and set to default values if no file exists.

void GetLastReboot(LARGE_INTEGER& LastReboot)
{

	HANDLE FileHandle = CreateFile
	(
		CFileName,
		GENERIC_READ,
		0,
		0,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
		0
	);

	if(FileHandle == INVALID_HANDLE_VALUE)
	{
		// File does not (yet) exist; set LastReboot to def.values.
		LastReboot.QuadPart = CDefSysTime;

		return;
	}

	DWORD BytesRead;

	// Read struct from file.
	ReadFile
	(
		FileHandle,
		&LastReboot,
		sizeof(LARGE_INTEGER),
		&BytesRead,
		0
	);
	

	CloseHandle(FileHandle);

}

//---------------------------------------------------------------------------

// Return true if we're online, else false.

bool IsOnline()
{

	WSADATA WSAData;

	if(WSAStartup(MAKEWORD(2, 1), &WSAData) != 0)
	{
		return false;
	}

	SOCKET SockRaw = WSASocket
	(
		AF_INET,
		SOCK_RAW,
		IPPROTO_ICMP,
		0,
		0,
		0
	);
  
	if(SockRaw == INVALID_SOCKET)
	{
		return false;
	}

	int Timeout = CSocketTimeout;

	int BRead;

	BRead = setsockopt
	(
		SockRaw,
		SOL_SOCKET,
		SO_RCVTIMEO,
		reinterpret_cast <char*> (&Timeout),
		sizeof(Timeout)
	);

	if(BRead == SOCKET_ERROR)
	{
		return false;
	}

	sockaddr_in Destination;

	memset(&Destination, 0, sizeof(Destination));

	Destination.sin_addr.s_addr = inet_addr(CTargetAddress);
	Destination.sin_family      = AF_INET;

	int DataSize = CDefPacketSize;
	
	DataSize += sizeof(TICMPHeader);  

	char* ICMPData = new char[CMaxPacket];

	if(!ICMPData)
	{
		return false;
	}
  
	memset(ICMPData, 0, CMaxPacket);

	FillIcmpData(ICMPData, DataSize);

	((TICMPHeader*)ICMPData)->TimeStamp = GetTickCount();
	((TICMPHeader*)ICMPData)->Sequence  = 0;
	((TICMPHeader*)ICMPData)->Checksum  = Checksum((USHORT*)ICMPData, DataSize);

	int BWrote = sendto
	(
		SockRaw,
		ICMPData,
		DataSize,
		0,
		reinterpret_cast <sockaddr*> (&Destination),
		sizeof(Destination)
	);

	if(BWrote == SOCKET_ERROR)
	{
		delete [] ICMPData;

		return false;
	}

	char* ReceiveBuffer = new char[CMaxPacket];
	if(!ReceiveBuffer)
	{
		delete [] ICMPData;

		return false;
	}

	sockaddr_in From;
	int FromLength = sizeof(From);

	BRead = recvfrom
	(
		SockRaw,
		ReceiveBuffer,
		CMaxPacket,
		0,
		reinterpret_cast <sockaddr*> (&From),
		&FromLength
	);

	delete [] ReceiveBuffer;
	delete [] ICMPData;

	if(BRead == SOCKET_ERROR)
	{
		return false;
	}

	// We're online!
	return true; 

}

//---------------------------------------------------------------------------

// Kill some tasks that need a proper byebye before we hit the cold reset.

void KillTasks()
{

	static const int AmountOfTasksToKill = 3;

	// Entry "WinVNC" removed; 07092001.

	static const char TasksToKill[AmountOfTasksToKill][32] =
	{
		"BulletProof FTP Server v2.15",
		"Engine Status",
		"Tsp3216LWindow"
	};

	for(int Idx = 0; Idx < AmountOfTasksToKill; ++Idx)
	{
		// Hit every task with three messages
		// to simulate a logoff/reboot.

		HWND Window = FindWindow(0, TasksToKill[Idx]);

		if(IsWindow(Window))
		{
			PostMessage
			(
				Window,
				WM_QUERYENDSESSION,
				0,
				ENDSESSION_LOGOFF
			);

			PostMessage
			(
				Window,
				WM_ENDSESSION,
				1,
				ENDSESSION_LOGOFF
			);

			PostMessage
			(
				Window,
				WM_QUIT,
				0,
				0
			);
		}
	}

}

//---------------------------------------------------------------------------


LRESULT CALLBACK MessageHandler
	(
		HWND Window,
		UINT Message,
		WPARAM WParam,
		LPARAM LParam
	)
{

	static int Counter = 0;

	switch(Message)
	{
		case WM_DESTROY:
			PostQuitMessage(0);

			return 0;

		case WM_TIMER:
			++Counter;

			switch(MMode)
			{
				case emNeutral:
					if(Counter >= CCheckInterval)
					{
						Counter = 0;
						DoCheck(MMode);
					}

					#ifndef NODEBUG
					ShowStatus(Window, MMode, Counter, CCheckInterval);
					#endif

					break;

				case emArmedAndDangerous:
					if(Counter >= CReCheckInterval)
					{
						Counter = 0;
						DoCheck(MMode);
					}

					#ifndef NODEBUG
					ShowStatus(Window, MMode, Counter, CReCheckInterval);
					#endif

					break;

				case emFinished:
					#ifndef NODEBUG
					ShowStatus(Window, MMode, 0, 0);
					#endif

					break;

				// default case skipped on purpose.
			}

			return 0;


		// default case skipped on purpose.
	}

	return DefWindowProc(Window, Message, WParam, LParam);

}

//---------------------------------------------------------------------------

// Returns true if recently rebooted, in which case we will not even
// check if we're online or not. Use 64 bit maths.

bool RecentlyRebooted()
{

	// What time do we have now?
	LARGE_INTEGER Now;
	GetSystemTimeAsFileTime(reinterpret_cast <FILETIME*> (&Now));

	return (Now.QuadPart - MLastReboot.QuadPart) < CMinRebootDiff;

}

//---------------------------------------------------------------------------

// Turn relais on or off.

void Relais(const bool On)
{

	UCHAR KeyboardState[256];
	GetKeyboardState(KeyboardState);
	KeyboardState[VK_NUMLOCK] = On ? 1 : 0;
	SetKeyboardState(KeyboardState);

}


//---------------------------------------------------------------------------

// Show current status.

#ifndef NODEBUG

void ShowStatus
	(
		HWND Window, 
		const TMode CurrentMode,
		const int CurrentTick,
		const int TotalTicks
	)
{

	char Buffer[256];

	SYSTEMTIME SystemTime;

	FileTimeToSystemTime
	(
		reinterpret_cast <FILETIME*> (&MLastReboot),
		&SystemTime
	);

	wsprintf
	(
		Buffer, 
		"%d %d - Last reboot: %.02d %.02d %.04d %.02d:%.02d:%.02d ", 
		CurrentTick, 
		TotalTicks,
		SystemTime.wMonth,
		SystemTime.wDay,
		SystemTime.wYear,
		SystemTime.wHour,
		SystemTime.wMinute,
		SystemTime.wSecond
	);

	switch(CurrentMode)
	{
		case emNeutral:
			lstrcat(Buffer, "neutral");
			break;

		case emArmedAndDangerous:
			lstrcat(Buffer, "armed and dangerous!");
			break;

		case emFinished:
			lstrcat(Buffer, "hasta la vista, baby...");
			break;
	}

	SetWindowText(Window, Buffer);

}

#endif

//---------------------------------------------------------------------------

int APIENTRY WinMain(HINSTANCE Instance, HINSTANCE, LPSTR, int)
{

	// Turn off NumLock just in case.
	Relais(false);

	// Create an invisible window.
	WNDCLASS WindowClass;

	WindowClass.cbClsExtra    = 0;
	WindowClass.cbWndExtra    = 0;
	WindowClass.hbrBackground = reinterpret_cast <HBRUSH>
	                            (GetStockObject(BLACK_BRUSH));
	WindowClass.hCursor       = LoadCursor(0, IDC_ARROW);
	WindowClass.hIcon         = LoadIcon(0, IDI_APPLICATION);
	WindowClass.hInstance     = Instance;
	WindowClass.lpfnWndProc   = MessageHandler;
	WindowClass.lpszClassName = CAppName;
	WindowClass.lpszMenuName  = 0;
	WindowClass.style         = 0;

	if(!RegisterClass(&WindowClass))
	{
		return 0;
	}

	#ifdef NODEBUG
	DWORD WindowStyle = WS_OVERLAPPED;
	#else
	DWORD WindowStyle = WS_OVERLAPPEDWINDOW;
	#endif

	HWND Window = CreateWindow
	(
		CAppName,
		CAppName,
		WindowStyle,
		0,
		0,
		400,
		0,
		0,
		0,
		Instance,
		0
	);

	if(!IsWindow(Window))
	{
		return 0;
	}
	
	#ifdef NODEBUG
	ShowWindow(Window, SW_HIDE);
	#else
	ShowWindow(Window, SW_SHOWMINIMIZED);
	#endif

	// Read last reboot time from file so we can compare.
	GetLastReboot(MLastReboot);

	// Start a timer.
	UINT TimerID = SetTimer
	(
		Window,
		1,
		CTimerInterval,
		0
	);

	// Start message pump.
	MSG Message;

	while(GetMessage(&Message, 0, 0, 0))
	{
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}

	KillTimer(Window, TimerID);

	return 0;
}


//---------------------------------------------------------------------------

// Update last reboot for both input parameter and file to this moment.

void UpdateLastReboot(LARGE_INTEGER& LastReboot)
{

	// What time do we have now?
	FILETIME Now;
	GetSystemTimeAsFileTime(&Now);

	// Update input parameter.
	memcpy(&LastReboot, &Now, sizeof(FILETIME));

	// Update time in file.
	HANDLE FileHandle = CreateFile
	(
		CFileName,
		GENERIC_WRITE,
		0,
		0,
		CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,
		0
	);

	if(FileHandle != INVALID_HANDLE_VALUE)
	{
		DWORD BytesWritten;

		// Write struct to file.
		WriteFile
		(
			FileHandle,
			&Now,
			sizeof(FILETIME),
			&BytesWritten,
			0
		);
	
		CloseHandle(FileHandle);
	}

}

//---------------------------------------------------------------------------
